"""Create a colored diff."""

import difflib


def inserted(new_text: str) -> str:
    """Return the annotated text."""
    return f"[inserted]{new_text}[/inserted]"


def deleted(old_text: str) -> str:
    """Return the annotated text."""
    return f"[deleted]{old_text}[/deleted]"


def colored_diff(old_text: str, new_text: str) -> str:
    """Return a colored string showing the diffs between old and new text."""
    matcher = difflib.SequenceMatcher(a=old_text.lower(), b=new_text.lower())
    if matcher.ratio() < 0.6:
        return inserted(new_text)
    result = ""
    for operator, old_start, old_end, new_start, new_end in matcher.get_opcodes():
        old_fragment, new_fragment = old_text[old_start:old_end], new_text[new_start:new_end]
        if operator == "delete":
            result += deleted(old_fragment)
        elif operator == "insert":
            result += inserted(new_fragment)
        elif operator == "replace":
            if len(old_fragment) == len(new_fragment) == 1:
                result += (deleted(old_fragment) + inserted(new_fragment))
            else:
                result += inserted(new_fragment)
        else:
            result += new_fragment
    return result



"""Meta data about this application."""

import pathlib
from importlib.metadata import metadata, version
from typing import Literal


_metadata = metadata("Toisto")
NAME = _metadata["name"]
SUMMARY = _metadata["summary"]

VERSION = version(NAME)
SUPPORTED_LANGUAGES = dict(en="English", fi="Finnish", nl="Dutch")
Language = Literal["en", "fi", "nl"]

_topics_folder = pathlib.Path(__file__).parent / "topics"
TOPIC_JSON_FILES = list(_topics_folder.glob("*.json"))
TOPICS = sorted(json_file.stem for json_file in TOPIC_JSON_FILES)



"""Text matching."""

import string


def without_punctuation(text: str) -> str:
    """Remove text without punctuation."""
    return ''.join(char for char in text if char not in string.punctuation)


def match(text1: str, *texts: str) -> bool:
    """Return whether the text matches any of the texts."""
    return any(
        without_punctuation(text1.strip().lower()) == without_punctuation(text2.strip().lower()) for text2 in texts
    )



"""Classes for storing and loading data."""

import errno
import json
import os
import pathlib

from .metadata import NAME, TOPIC_JSON_FILES, Language
from .model import concept_factory, Progress, Quiz
from .output import show_error_and_exit


PROGRESS_JSON = pathlib.Path.home() / f".{NAME.lower()}-progress.json"


def load_json(json_file_path: pathlib.Path, default=None):
    """Load the JSON from the file. Return default if file does not exist."""
    if json_file_path.exists():
        with json_file_path.open(encoding="utf-8") as json_file:
            return json.load(json_file)
    if default is None:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(json_file_path))
    return default


def dump_json(json_file_path: pathlib.Path, contents) -> None:
    """Dump the JSON into the file."""
    with json_file_path.open("w", encoding="utf-8") as json_file:
        json.dump(contents, json_file)


def load_quizzes(
    language: Language, source_language: Language, topics_to_load: list[str], topic_files_to_load: list[str]
) -> list[Quiz]:
    """Load the entries from the topics and generate the quizzes."""
    quizzes = []
    topics: list[pathlib.Path] = []
    if topics_to_load or topic_files_to_load:
        topics.extend(topic for topic in TOPIC_JSON_FILES if topic.stem in topics_to_load)
        topics.extend(pathlib.Path(topic) for topic in topic_files_to_load)
    else:
        topics.extend(TOPIC_JSON_FILES)
    for topic in topics:
        try:
            for concept_dict in load_json(topic):
                concept = concept_factory(concept_dict)
                quizzes.extend(concept.quizzes(language, source_language))
        except Exception as reason:  # pylint: disable=broad-except
            show_error_and_exit(f"{NAME} cannot read topic {topic}: {reason}.\n")
    return quizzes


def load_progress() -> Progress:
    """Load the progress from the user's home folder."""
    try:
        progress_dict = load_json(PROGRESS_JSON, default={})
        return Progress(progress_dict)
    except Exception as reason:  # pylint: disable=broad-except
        show_error_and_exit(
            f"""{NAME} cannot parse the progress information in {PROGRESS_JSON}: {reason}.
To fix this, remove or rename {PROGRESS_JSON} and start {NAME} again. Unfortunately, this will reset your progress.
Please consider opening a bug report at https://github.com/fniessink/{NAME.lower()}. Be sure to attach the invalid
progress file to the issue.
""")


def save_progress(progress: Progress) -> None:
    """Save the progress to the user's home folder."""
    dump_json(PROGRESS_JSON, progress.as_dict())



"""Concept classes."""

from dataclasses import dataclass
from typing import cast, Literal

from toisto.metadata import Language

from .label import Labels
from .quiz import Quiz


ConceptDict = dict[Language, str | list[str]]
NounType = Literal["plural", "singular"]
NounConceptDict = dict[NounType, ConceptDict]


@dataclass
class Concept:
    """Class representing a concept from a topic."""

    _labels: dict[Language, Labels]

    def quizzes(self, language: Language, source_language: Language) -> list[Quiz]:
        """Generate the possible quizzes from the concept and its labels."""
        return (
            [Quiz(language, source_language, label, self.labels(source_language)) for label in self.labels(language)] +
            [Quiz(source_language, language, label, self.labels(language)) for label in self.labels(source_language)]
        ) if self.has_labels(language, source_language) else []

    def has_labels(self, *languages: Language) -> bool:
        """Return whether the concept has labels for all the specified languages."""
        return all(language in self._labels for language in languages)

    def labels(self, language: Language) -> Labels:
        """Return the labels for the language."""
        return self._labels[language]

    @classmethod
    def from_dict(cls, concept_dict: ConceptDict) -> "Concept":
        """Instantiate a concept from a dict."""
        return cls(
            {
                language: cast(Labels, (label if isinstance(label, list) else [label]))
                for language, label in concept_dict.items()
            }
        )


@dataclass
class NounConcept:
    """A concept that is a noun with a singular and plural grammatical number.

    See https://en.wikipedia.org/wiki/Grammatical_number.
    """

    singular: Concept
    plural: Concept

    def quizzes(self, language: Language, source_language: Language) -> list[Quiz]:
        """Generate the possible quizzes from the concept."""
        result = []
        for concept in self.singular, self.plural:
            if concept.has_labels(language, source_language):
                result.extend(concept.quizzes(language, source_language))
        if self.singular.has_labels(language) and self.plural.has_labels(language):
            singular_labels, plural_labels = self.singular.labels(language), self.plural.labels(language)
            result.extend([Quiz(language, language, label, plural_labels, "pluralize") for label in singular_labels])
            result.extend([Quiz(language, language, label, singular_labels, "singularize") for label in plural_labels])
        return result

    @classmethod
    def from_dict(cls, concept_dict: NounConceptDict) -> "NounConcept":
        """Instantiate a concept from a dict."""
        singular = Concept.from_dict(concept_dict["singular"])
        plural = Concept.from_dict(concept_dict["plural"])
        return cls(singular, plural)


def concept_factory(concept_dict: ConceptDict | NounConceptDict) -> Concept | NounConcept:
    """Create a concept from the concept dict."""
    if "singular" in concept_dict and "plural" in concept_dict:
        return NounConcept.from_dict(cast(NounConceptDict, concept_dict))
    return Concept.from_dict(cast(ConceptDict, concept_dict))



"""Progress model class."""

import random

from .quiz import Quiz
from .quiz_progress import QuizProgress


class Progress:
    """Keep track of progress on quizzes."""
    def __init__(self, progress_dict: dict[str, dict[str, int | str]]) -> None:
        self.progress_dict = {key: QuizProgress.from_dict(value) for key, value in progress_dict.items()}
        self.current_quiz: Quiz | None = None

    def update(self, quiz: Quiz, correct: bool) -> None:
        """Update the progress on the quiz."""
        key = str(quiz)
        self.progress_dict.setdefault(key, QuizProgress()).update(correct)

    def get_progress(self, quiz: Quiz) -> QuizProgress:
        """Return the progress on the quiz."""
        key = str(quiz)
        return self.progress_dict.get(key, QuizProgress())

    def next_quiz(self, quizzes: list[Quiz]) -> Quiz | None:
        """Return the next quiz."""
        if eligible_quizzes := [quiz for quiz in quizzes if self.is_eligible(quiz)]:
            if len(eligible_quizzes) > 1 and self.current_quiz in eligible_quizzes:
                eligible_quizzes.remove(self.current_quiz)
            self.current_quiz = random.choice(eligible_quizzes)
            return self.current_quiz
        return None
#
    def is_eligible(self, quiz: Quiz) -> bool:
        """Is the quiz eligible?"""
        return not self.get_progress(quiz).is_silenced()

    def as_dict(self) -> dict[str, dict[str, int | str]]:
        """Return the progress as dict."""
        return {key: value.as_dict() for key, value in self.progress_dict.items()}



"""Quiz classes."""

from dataclasses import dataclass
from typing import Literal

from toisto.match import match
from toisto.metadata import Language, SUPPORTED_LANGUAGES

from .label import Label, Labels


QuizType = Literal["translate", "pluralize", "singularize"]
INSTRUCTION: dict[QuizType, str] = dict(
    translate="Translate into",
    pluralize="Give the [underline]plural[/underline] in",
    singularize="Give the [underline]singular[/underline] in"
)


@dataclass
class Quiz:
    """Class representing a quiz."""
    question_language: Language
    answer_language: Language
    question: Label
    answers: Labels
    quiz_type: QuizType = "translate"

    def is_correct(self, guess: Label) -> bool:
        """Return whether the guess is correct."""
        return match(guess, *self.answers)

    def get_answer(self) -> Label:
        """Return the first answer."""
        return self.answers[0]

    def other_answers(self, guess: Label) -> Labels:
        """Return the answers not equal to the guess."""
        assert self.is_correct(guess)
        return [answer for answer in self.answers if not match(guess, answer)]

    def instruction(self) -> str:
        """Generate the quiz instruction."""
        return f"{INSTRUCTION[self.quiz_type]} {SUPPORTED_LANGUAGES[self.answer_language]}"



"""Practice command."""

from toisto.model import Label, Progress, Quiz
from toisto.output import console, feedback_correct, feedback_incorrect, instruction, DONE, WELCOME, TRY_AGAIN
from toisto.persistence import save_progress
from toisto.speech import say


def do_quiz(quiz: Quiz, progress: Progress) -> None:
    """Do one quiz and update the progress."""
    console.print(instruction(quiz))
    say(quiz.question_language, quiz.question)
    guess = Label(input("> "))
    correct = quiz.is_correct(guess)
    if not correct:
        console.print(TRY_AGAIN)
        guess = Label(input("> "))
        correct = quiz.is_correct(guess)
    progress.update(quiz, correct)
    quiz_progress = progress.get_progress(quiz)
    console.print(feedback_correct(guess, quiz, quiz_progress) if correct else feedback_incorrect(guess, quiz))


def practice(quizzes, progress: Progress) -> None:
    """Practice a language."""
    console.print(WELCOME)
    try:
        while quiz := progress.next_quiz(quizzes):
            do_quiz(quiz, progress)
        console.print(DONE)
    except (KeyboardInterrupt, EOFError):
        console.print()  # Make sure the shell prompt is displayed on a new line
    finally:
        save_progress(progress)
